class TestClass:
    def __init__(self):
        pass

    def printHello(self):
        print('Hello')

class TestClass2:
    def __init__(self):
        pass

    def printHello(self):
        print('Hello')


# Example code
if __name__ == '__main__':
    print("This should not be printed!!")